@@ -1,7 +1,7 @@
 """Synapse Query Store
 """
 import pandas as pd
-from schematic_db.synapse import Synapse, SynapseConfig
+from schematic_db.synapse.synapse import Synapse
 from .query_store import QueryStore
 
 
@@ -11,12 +11,13 @@ class SynapseQueryStore(QueryStore):  # pylint: disable=too-few-public-methods
     - An adaptor between Synapse class and QueryStore ABC
     """
 
-    def __init__(self, config: SynapseConfig):
+    def __init__(self, auth_token: str, project_id: str):
         """Init
         Args:
-            config (SynapseConfig): A SynapseConfig object
+            auth_token (str): A Synapse auth_token
+            project_id (str): A Synapse id for a project
         """
-        self.synapse = Synapse(config)
+        self.synapse = Synapse(auth_token, project_id)
 
     def store_query_result(self, table_name: str, query_result: pd.DataFrame) -> None:
         self.synapse.replace_table(table_name, query_result)
